import os
import logging
import cStringIO
from xml.dom import minidom
from xml.parsers.expat import ExpatError
from xml.sax.handler import ContentHandler

import pmm_config
import pmm_dump_formatter
import migration_handler_rc
import pmm_dump_access_service
import migration_handler_session
from pmmcli_exceptions import PMMUtilityException
from pmm_api_xml_protocols import DumpsStorageCredentials, DumpSpecification, \
RestoreTaskDescription, RestoreTaskDescriptionSource, RestoreTaskDescriptionObjects, \
RestoreTaskDescriptionMisc, Rule, Resolution, ResolutionRename, Response, SessionId, \
ResolutionOverwrite, TimingPolicy, Policy, ConflictResolutionRules, \
ResolveConflictsTaskDescription, IgnoreError, IgnoreErrors

from encoder import EncoderFile
from subproc import NonzeroExitException


_logger = logging.getLogger("pmm.migration_handler.restore")


class MigrationRestoreException(Exception):
    def __init__(self, message):
        Exception.__init__(self, message)
        self.msg = message

    def get_message(self):
        return self.msg


class XmlElementOwnerGuidAttrSetter(pmm_dump_access_service.XmlWorker):
    def __init__(self, owners_mapping):
        self.__owners_mapping = owners_mapping
        self.__level = -1

    def startElement(self, name, attrs):
        if name == "migration-dump":
            self.__level = 0
        else:
            self.__level += 1
        if self.__level == 1:
            for setOwner in self.__owners_mapping.get_set_owner():
                for child in setOwner.get_child():
                    child_name = child.get_name()
                    attrs_name = None
                    if attrs.has_key("name"):
                        attrs_name = attrs.getValue("name")
                    owner_guid = setOwner.get_owner().get_guid()
                    if child.get_type() == name and ( child_name is None or child_name in ['', attrs_name] ):
                        _logger.debug("Map %s with name '%s' to owner %s" % (name, attrs_name, owner_guid))
                        attrs._attrs["owner-guid"] = owner_guid

    def endElement(self, name, value):
        self.__level -= 1


class XmlElementOwnerGuidAttrResetter(pmm_dump_access_service.XmlWorker):
    def __init__(self):
        self.__level = -1

    def startElement(self, name, attrs):
        if name == "migration-dump":
            self.__level = 0
        else:
            self.__level += 1
        if self.__level == 1:
            if attrs.has_key("owner-guid"):
                attrs._attrs["owner-guid"] = ""

    def endElement(self, name, value):
        self.__level -= 1


class XMLFilterSimpleWriter(pmm_dump_access_service.XMLSimpleWriter):
    def __init__(self):
        pmm_dump_access_service.XMLSimpleWriter.__init__(self)

    def setSubfilters(self, subfilter_stack):
        pass


#
# Base class for XmlFilterWorker-derived classes
#
class XmlFilterWorker:
    def __init__(self):
        self.__subfilter = None

    def setSubfilters(self, subfilter_stack = []):
        if len(subfilter_stack) > 0:
            self.__subfilter = subfilter_stack[0]
            self.__subfilter.setSubfilters(subfilter_stack[1:])

    def getSubfilter(self):
        return self.__subfilter

    def startElement(self, name, attrs):
        if self.__subfilter is not None:
            self.__subfilter.startElement(name, attrs)

    def endElement(self, name, value):
        if self.__subfilter is not None:
            self.__subfilter.endElement(name, value)


class DBServersMapper(XmlFilterWorker):
    def __init__(self, db_mapping):
        XmlFilterWorker.__init__(self)
        self.__db_mapping = db_mapping
        self.__in_dbserver_xml = False
        self.__dbserver_xml = {}

    def startElement(self, name, attrs):
        if name == "db-server":
            self.__in_dbserver_xml = True
        if self.__in_dbserver_xml:
            self.__dbserver_xml[name] = {"attrs": attrs}
        else:
            self.getSubfilter().startElement(name, attrs)

    def endElement(self, name, value):
        if self.__in_dbserver_xml:
            self.__dbserver_xml[name]["value"] = value
            if name == "db-server":
                self.__mapDbServer()
                self.__in_dbserver_xml = False
                self.__dbserver_xml = {}
        else:
            self.getSubfilter().endElement(name, value)

    def __mapDbServer(self):
        # try find mapping for self.__db_server_stack
        destination_object = None
        for mapping in self.__db_mapping.get_mapping():
            source_object = mapping.get_source()
            if source_object.get_type() == self.__dbserver_xml["db-server"]["attrs"].getValue("type") and \
               source_object.get_host() == self.__dbserver_xml["host"]["value"] and \
               source_object.get_port() == self.__dbserver_xml["port"]["value"]:
                destination_object = mapping.get_destination()
                break
        
        if destination_object is not None:
            # found destination_object for mapping
            subfilter = self.getSubfilter()
            # do mapping
            self.__dbserver_xml["db-server"]["attrs"]._attrs["type"] = destination_object.get_type()
            self.__dbserver_xml["host"]["value"] = destination_object.get_host()
            self.__dbserver_xml["port"]["value"] = destination_object.get_port()
            keys = self.__dbserver_xml.keys()
            subfilter.startElement( "db-server", self.__dbserver_xml["db-server"]["attrs"])
            subfilter.startElement( "host", self.__dbserver_xml["host"]["attrs"])
            subfilter.endElement( "host", self.__dbserver_xml["host"]["value"])
            subfilter.startElement( "port", self.__dbserver_xml["port"]["attrs"])
            subfilter.endElement( "port", self.__dbserver_xml["port"]["value"])
            if "db-admin" in keys:
                subfilter.startElement( "db-admin", self.__dbserver_xml["db-admin"]["attrs"])
                subfilter.startElement( "password", self.__dbserver_xml["password"]["attrs"])
                subfilter.endElement( "password", self.__dbserver_xml["password"]["value"])
                subfilter.endElement( "db-admin", self.__dbserver_xml["db-admin"]["value"])
            subfilter.endElement( "db-server", self.__dbserver_xml["db-server"]["value"])


class XmlFilterHandler(ContentHandler):
    def __init__ (self, filter_stack = []):
        ContentHandler.__init__(self)
        self.__filter_stack = filter_stack
        self.__filter = None
        if len(self.__filter_stack) > 0:
            self.__filter = self.__filter_stack[0]
            self.__filter.setSubfilters(self.__filter_stack[1:])
        self.__elementValue = []

    def startElement(self, name, attrs):
        self.__elementValue.append(u'')
        if self.__filter is not None:
            self.__filter.startElement(name, attrs)

    def characters (self, ch):
        self.__elementValue[-1] = self.__elementValue[-1] + ch

    def endElement(self, name):
        if self.__filter is not None:
            self.__filter.endElement(name, self.__elementValue[-1])
        self.__elementValue.pop()


def dbServersEquals(dbServer1, dbServer2):
    if not valuesEqual(dbServer1.get_type(), dbServer2.get_type()):
        return False
    if not valuesEqual(dbServer1.get_host(), dbServer2.get_host()):
        return False
    if not valuesEqual(dbServer1.get_port(), dbServer2.get_port()):
        return False
    return True


def valuesEqual(obj1, obj2):
    if obj1 is None:
        return obj2 is None
    if obj2 is None:
        return False
    return obj1 == obj2


def dbServerToString(dbServer):
    if dbServer is None:
        return 'None'
    dbStrArray = []
    host = dbServer.get_host()
    if host is not None:
        dbStrArray.append("host:%s " % host)
    port = dbServer.get_port()
    if port is not None:
        dbStrArray.append("port:%s " % port)
    dbtype = dbServer.get_type()
    if dbtype is not None:
        dbStrArray.append("type:%s " % dbtype)
    return ''.join(dbStrArray)


# returns an array of conflict resolution rules (pmm_api_xml_protocols.Rule() instances)
def create_resolution_rules_from_maps(conflicts_description, ipMap):
    _logger.debug("Generate ip and db mapping for conflict resolution")
    #Create ip map conflict resolution
    conflicts = conflicts_description.get_conflict()
    if len(conflicts) == 0:
        _logger.warning("No conflicts found!")
        return []
    
    pmm_resolution_rules = []
    for conflict in conflicts:
        conflict_type = conflict.get_type() 
        if conflict_type is not None:
            conflict_type_configuration = conflict_type.get_configuration() 
            if conflict_type_configuration is not None:
                #Configuration conflict
                configuration = conflict.get_type().get_configuration()
                if ipMap is not None:
                    ip = configuration.get_reason_description().get_required_resource_description().get_ip()
                    if ip is not None:
                        _logger.debug("Prepare ip map conflict for ip:%s" % ip.get_value())
                        iptype = None
                        ipdest = None
                        ipMapShared =ipMap.get_shared() 
                        if ipMapShared is not None:
                            for sharedip in ipMapShared:
                                if sharedip.get_source() == ip.get_value():
                                    iptype = "shared"
                                    ipdest = sharedip.get_destination()
                        if ipdest is None:
                            ipMapExclusive = ipMap.get_exclusive() 
                            if ipMapExclusive is not None:
                                for exclusiveip in ipMapExclusive:
                                    if exclusiveip.get_source() == ip.get_value():
                                        iptype = "exclusive"
                                        ipdest = exclusiveip.get_destination()
                        if ipdest is not None:
                            #add conflict resolution rule
                            _logger.debug("Map conflict for ip:%s to %s:%s" % (ip.get_value(),iptype,ipdest))
                            rule = Rule()
                            rule.set_conflict_id(conflict.get_id())
                            rule.set_resolution(Resolution())
                            resolution_rename_object = ResolutionRename()
                            resolution_rename_object.setValueOf_('')
                            rule.get_resolution().set_rename(resolution_rename_object)
                            rule.get_resolution().get_rename().set_new_name("ip-type:" + iptype + ":ip-address:" + ipdest)
                            pmm_resolution_rules.append(rule)
                        else:
                            _logger.error("Cannot find mapping for ip:" + ip.get_value() )
    return pmm_resolution_rules


def getDefaultRestorePolicy():
    _logger.debug("default restore policy generated")
    pmm_resolution_rules_policy = Policy()
    return pmm_resolution_rules_policy


def export_object(object_to_export, object_name):
    object_str = cStringIO.StringIO()
    object_encoder = EncoderFile(object_str, 'utf-8')
    object_encoder.write('<?xml version="1.0" encoding="UTF-8"?>\n')
    object_to_export.export(object_encoder, 0, name_ = object_name)
    return object_str.getvalue()


def store_versioned_file(dir, filename, content):
    file_name_base = os.path.join(dir, filename)
    file_name = file_name_base
    index = 0
    while os.path.exists(file_name):
        index += 1
        file_name = file_name_base + "." + str(index)
    
    _logger.debug("Saving file %s ..." % file_name)
    file_object = open(file_name,'wt')
    file_object.write(content)
    file_object.close()
    _logger.debug("Saving file %s done" % file_name)
    
    return file_name


class MigrationRestore:
    def __init__(self, migration_session_id):
        self.__migration_session_id = migration_session_id
        self.__migration_session = migration_handler_session.MigrationSession(migration_session_id)
        
        self.__last_session_id = None
        self.__conflicts_description = None
        self.__dump_specification_formatter = None

    def __createDumpSpecificationFormatter(self):
        # get dump file name from session directory ('dump-name')
        # create dump_specification_formatter
        dump_file_name = self.__migration_session.getValue('dump-name')
        if dump_file_name is not None:
            if not os.path.isfile(dump_file_name):
                raise MigrationRestoreException("Dump file name is not exists: %s" % dump_file_name)
        
        dump_file_name_head, dump_file_name_tail = os.path.split(dump_file_name)
        
        dumps_storage_credentials_object = DumpsStorageCredentials()
        dumps_storage_credentials_object.set_storage_type('local')
        dumps_storage_credentials_object.set_root_dir(dump_file_name_head)
        dump_specification_object = DumpSpecification(dumps_storage_credentials = dumps_storage_credentials_object)
        dump_specification_object.set_name_of_info_xml_file(dump_file_name_tail)
        dump_specification_formatter = pmm_dump_formatter.DumpSpecificationFormatter(dump_specification_object)
        return dump_specification_formatter

    def __prepareCmdLine(self, filename):
        return self.__migration_session.createWorker(filename)

    def __runCmdLine(self, cmd):
        self.__migration_session.runWorker(cmd)

    def __mapDBServers(self, db_mapping):
        _logger.debug("Start DBServers mapping")
        dump_name = self.__migration_session.getValue('dump-name')
        writer = XMLFilterSimpleWriter()
        dbmapper = DBServersMapper(db_mapping)
        xml_processor = pmm_dump_access_service.InfoXMLProcessor(dump_name, XmlFilterHandler([dbmapper,writer]))
        xml_processor.process()
        newdumpname = self.__migration_session.getValue('dump-name') + ".dbmapped"
        content = open(newdumpname,'wt')
        content.write(writer.get_content())
        content.close()
        self.__migration_session.setValue('dump-name', newdumpname)
        _logger.debug("End DBServers mapping")

    def __mapOwners(self):
        owners_mapping = self.__migration_session.getObject('owner_mapping')
        writer = pmm_dump_access_service.XMLSimpleWriter()
        setter = None
        if owners_mapping is not None:
            _logger.debug("Run migration owner mapping" )
            setter = XmlElementOwnerGuidAttrSetter(owners_mapping)
        else:
            _logger.debug("Run migration owner resetting")
            setter = XmlElementOwnerGuidAttrResetter()
        xml_processor = pmm_dump_access_service.InfoXMLProcessor(self.__migration_session.getValue('dump-name'), pmm_dump_access_service.XmlHandler([setter,writer]))
        xml_processor.process()
        newdumpname = self.__migration_session.getValue('dump-name') + ".withowners"
        content = open(newdumpname,'wt')
        content.write(writer.get_content())
        content.close()
        self.__migration_session.setValue('dump-name', newdumpname)
        _logger.debug("Dump name changed due to owners mapping: '%s'" % newdumpname)

    def __invokeResolveConflicts(self, resolve_conflicts_task_description_object):
        _logger.debug("__invokeResolveConflicts start")
        cmd = self.__prepareCmdLine(pmm_config.pmmcli())
        cmd.arg("--resolve-conflicts")
        cmd.stdin = export_object(resolve_conflicts_task_description_object, 'resolve-conflicts-task-description')
        
        try:
            self.__runCmdLine(cmd)
        except NonzeroExitException, x:
            raise PMMUtilityException('PMMCli resolve-conflicts execution failed with return code:' + str(x.exitcode), x)
        
        resolve_conflicts_response_file_name = store_versioned_file(self.__migration_session.getSessionPath(),'resolve_conflicts',cmd.get_stdout().encode('utf-8'))
        
        response = Response.factory()
        response.build(minidom.parse(resolve_conflicts_response_file_name).childNodes[0] )
        
        if response.get_errcode() != '0':
            raise MigrationRestoreException("PMMCli response contains errorcode '" + str(response.get_errcode()) + "'" )
        
        response_data_object = response.get_data()
        if response_data_object is None:
            raise MigrationRestoreException("PMMCli response is invalid. No data node")
        
        dump_overview_object = response_data_object.get_dump_overview()
        
        if dump_overview_object is None:
            raise MigrationRestoreException("PMMCli response is invalid. No dump-overview node")
        
        # store session_id
        session_id = dump_overview_object.get_session_id()
        _logger.debug("Save restore session:" + session_id)
        self.__last_session_id = session_id
        _logger.debug("__invokeResolveConflicts end")

    def __invokeRestore(self, restore_task_description_object):
        _logger.debug("__invokeRestore start")
        cmd = self.__prepareCmdLine(pmm_config.pmmcli())
        cmd.arg("--restore")
        cmd.stdin = export_object(restore_task_description_object, 'restore-task-description')
        try:
            self.__runCmdLine(cmd)
        except NonzeroExitException, x:
            raise PMMUtilityException('PMMCli restore execution failed with return code:' + str(x.exitcode), x)
        
        restore_response_file_name = store_versioned_file(self.__migration_session.getSessionPath(),'restore_response',cmd.get_stdout().encode('utf-8'))
        response = Response.factory()
        try:
            response.build(minidom.parse(restore_response_file_name).childNodes[0])
        except ExpatError, x:
            errmsg = u"Wrong output from 'pmmcli --restore': Xml reading error: " + unicode(x)
            _logger.error("Exception raised while parsing pmmcli --restore output: '%s'" % restore_response_file_name)
            _logger.error(errmsg)
            raise MigrationRestoreException(errmsg)
        
        if response.get_errcode()!='0':
            if response.get_errcode() == '201':
                raise MigrationRestoreException('Unable to start Deployer because there is another Deployer run')
            else:
                raise MigrationRestoreException("PMMCli response contains errorcode '" + str(response.get_errcode()) + "'")
        
        response_data_object = response.get_data()
        if response_data_object is None:
            raise MigrationRestoreException("PMMCli response is invalid. No data node")
        
        response_data_task_id_object = response_data_object.get_task_id()
        response_data_task_id = None
        response_data_restore_task_result = None
        
        if response_data_task_id_object is not None:
            response_data_task_id = response_data_task_id_object.getValueOf_()
            _logger.debug("PMMCli response returned task-id: %s" % response_data_task_id)
            _logger.debug("__invokeRestore end: start restore")
            return response_data_task_id
        
        response_data_restore_task_result = response_data_object.get_restore_task_result()
        if response_data_restore_task_result is None:
            errmsg = "PMMCli response is invalid. Data node does not contain 'task-id' or 'restore-task-result' node"
            _logger.error(errmsg)
            raise MigrationRestoreException(errmsg)
        _logger.debug("PMMCli response returned restore-task-result")
        # store session_id and conflict_description
        conflicts_description_object = response_data_restore_task_result.get_conflicts_description()
        self.__conflicts_description = conflicts_description_object
        _logger.debug("Save conflicts description. Total conflicts: %d" % len(conflicts_description_object.get_conflict()))
        session_id = response_data_restore_task_result.get_dump_overview().get_session_id()
        self.__last_session_id = session_id
        _logger.debug("Save restore session:" + session_id)
        
        # return None to indicate restore was not started
        _logger.debug("__invokeRestore end: has conflicts")
        return None

    def __setRestoreTaskDescriptionResolutionRules(self, restore_task_description_object):
        conflict_resolution_rules_object = self.__getResolutionRules()
        #conflict_resolution_rules_object is ready
        restore_task_description_object.set_ignore_errors(None)
        restore_task_description_object.set_conflict_resolution_rules(conflict_resolution_rules_object)

    def __getResolutionRules(self):
        conflict_resolution_rules_object = self.__migration_session.getObject('restore_policy')
        # all rules should be deleted from stored resolution rules!
        # rules should be generated each time according current conflicts and ip_map and db_map specified for migration.
        if conflict_resolution_rules_object is None:
            _logger.warning("Conflicts resolution policies are absent in session")
            conflict_resolution_rules_object = ConflictResolutionRules( policy = getDefaultRestorePolicy() )
        else:
            conflict_resolution_rules_object.set_rule([])
        
        ipMap = self.__migration_session.getObject('ip_mapping')
        
        if ipMap is None:
            _logger.warning("IP mapping is not specified. Migration handler will try restore dump using restore policy only")
        else:
            rules = create_resolution_rules_from_maps(self.__conflicts_description, ipMap)
            conflict_resolution_rules_object.set_rule(rules)
        
        return conflict_resolution_rules_object

    # create skeleton for restore-task-description object.
    # one should finish creation with  __setResolutionRules() or __setIgnoreConflict()
    def __createRestoreTaskDescription(self):
        restore_task_description_object = RestoreTaskDescription()
        restore_task_description_object.set_owner_type('server')
        owner_guid = self.__migration_session.getValue('owner_guid')
        if owner_guid is None:
            _logger.warning('Migration session does not contain owner_guid parameter')
            owner_guid = ''
        restore_task_description_object.set_owner_guid(owner_guid)
        
        source_object = RestoreTaskDescriptionSource()
        if self.__last_session_id is not None:
            # continue work with last restore session
            session_id_object = SessionId()
            session_id_object.setValueOf_(self.__last_session_id)
            source_object.set_session_id(session_id_object)
            source_object.set_dump_specification(None)
        else:
            # begin new restore session from migrated dump
            dump_specification_object = self.__dump_specification_formatter.buildXml()
            source_object.set_dump_specification(dump_specification_object)
            source_object.set_session_id(None)
        restore_task_description_object.set_source(source_object)
        
        objects_object = RestoreTaskDescriptionObjects()
        objects_object.set_all('')
        objects_object.set_selected( None )
        restore_task_description_object.set_objects(objects_object)

        restore_task_description_misc_object = RestoreTaskDescriptionMisc()
        restore_task_description_misc_object.set_verbose_level('5')
        restore_task_description_misc_object.setValueOf_('')
        restore_task_description_object.set_misc(restore_task_description_misc_object)
        
        # restore_task_description creation is almost completed.
        # 'ignore-errors' and 'conflict-resolution-rules' are left to add 
        return restore_task_description_object

    def __setIgnoreConflict(self, restore_task_description_object):
        #Ignore further errors
        ignore_error_object = IgnoreError()
        ignore_error_object.set_type("conflicts")
        ignore_error_object.setValueOf_('')
        
        ignore_errors_object = IgnoreErrors()
        ignore_errors_object.set_ignore_error([ignore_error_object])
        
        restore_task_description_object.set_conflict_resolution_rules(None)
        restore_task_description_object.set_ignore_errors(ignore_errors_object)

    # create skeleton for restore-task-description object.
    # one should finish creation with  __setResolutionRules() or __setIgnoreConflict()
    def __createResolveConflictsTaskDescription(self):
        _logger.debug("Create new resolve conflicts task description started")
        resolve_conflicts_task_description_object = ResolveConflictsTaskDescription()
        
        session_id = self.__last_session_id
        if session_id is None:
            errmsg = "Migration restore requires session_id to be able to invoke pmmcli --resolve-confilcts"
            _logger.error(errmsg)
            raise MigrationRestoreException(errmsg)
        resolve_conflicts_task_description_object.set_session_id(session_id)
        
        conflict_resolution_rules_object = self.__getResolutionRules()
        #conflict_resolution_rules_object is ready
        
        resolve_conflicts_task_description_object.set_conflict_resolution_rules(conflict_resolution_rules_object)
        
        return resolve_conflicts_task_description_object

    def __doRestore(self, ignore_conflicts = False):
        # invoke pmmcli --restore WO conflict resolution rules
        # this will check for conflicts and start restore if there will no conflicts found
        restore_task_description_object = self.__createRestoreTaskDescription()
        if ignore_conflicts:
            self.__setIgnoreConflict(restore_task_description_object)
        response_data_task_id = self.__invokeRestore(restore_task_description_object)
        # response_data_task_id could be None
        # it is supposed that response_data_task_id is not None if being called with ignore_conflicts = True
        if ignore_conflicts and response_data_task_id is None:
            errmsg = "Migration restore was not started being called with ignore_conflicts options"
            _logger.error(errmsg)
            raise MigrationRestoreException(errmsg)
        return response_data_task_id

    def __doResolve(self):
        # invoke pmmcli --resolve-confilcts
        # then call __doRestore
        resolve_conflicts_task_description_object = self.__createResolveConflictsTaskDescription()
        self.__invokeResolveConflicts(resolve_conflicts_task_description_object)
        response_data_task_id = self.__doRestore(ignore_conflicts = False)
        if response_data_task_id is not None:
            return response_data_task_id

    def restore(self):
        dbMap = self.__migration_session.getObject('db_mapping')
        if dbMap is not None:
            self.__mapDBServers(dbMap)
        self.__mapOwners()
        self.__dump_specification_formatter = self.__createDumpSpecificationFormatter()
        response_data_task_id = self.__doRestore(ignore_conflicts = False)
        if response_data_task_id is not None:
            return response_data_task_id
        iteration = 0
        max_migration_conflict_resolves = migration_handler_rc.get().get_migration_max_conflict_resolves()
        while iteration < max_migration_conflict_resolves:
            response_data_task_id = self.__doResolve()
            if response_data_task_id is not None:
                return response_data_task_id
            iteration += 1
        # iterations are ended but conflicts still exist
        return self.__doRestore( ignore_conflicts = True )

