import os
import logging
from xml.dom import minidom
from xml.sax import make_parser, handler

import migration_handler_config
from pmm_migr_remote_service import MigrationRemoteService
from pmm_api_xml_protocols import ScoutResult, MigrationObjectList, ObjTypeReseller, ObjTypeClient, ObjTypeDomain, \
domains, clients, IPMapping, IpPairType, TaskStatus, Working, WorkingProgress, Starting


_logger = logging.getLogger("pmm.migration_handler.remote_service")


class MigrationRemoteServiceUnix(MigrationRemoteService):
    def __init__(self, host, port, login, passwd, session):
        MigrationRemoteService.__init__(self, host, port, login, passwd, session)

    def addContext(self, cmd):
        cmd.var('PSA_PASSWORD', self.passwd)
        cmd.arg('--host=' + self.host)
        if self.port is not None:
            cmd.arg('--port=' + self.port)
        cmd.arg('--login=' + self.login)
        cmd.arg('--session-path=' + self.getSessionPath())

    def sendScout(self, system_type = None):
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--send-scout')
        self.addContext(cmd)
        self.runCmdLine(cmd)

    def getScoutResult(self):
        scout_result_path = os.path.join(self.getSessionPath(), 'scout-result.xml')
        scout_result_file = open(scout_result_path, 'r')
        scout_rsult_str = scout_result_file.read()
        scout_result = ScoutResult.factory()
        scout_result.build(minidom.parseString(scout_rsult_str).childNodes[0])
        return scout_result

    def removeScout(self):
        pass

    def deployAgent(self, agent_name, agent_dir):
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--deploy-agent')
        self.addContext(cmd)
        cmd.arg('--agent-name=' + agent_name)
        cmd.arg('--agent-dir=' + agent_dir)
        self.runCmdLine(cmd)
        self.session.setFlag('agent_uploaded')

    def backupImpl(self, agent_name, agent_dir):
        self.session.setValue('step', 'dump')
        
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--dump-selected')
        self.addContext(cmd)
        cmd.arg('--selected-objects-file=' + self.session.getSelectedObjectsPath())
        cmd.arg('--agent-name=' + agent_name)
        cmd.arg('--agent-dir=' + agent_dir)
        self.runCmdLine(cmd)
        
        #download dump
        self.session.setValue('step','download')
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        self.addContext(cmd)
        cmd.arg('--get-content')
        cmd.arg('--agent-name=' + agent_name)
        cmd.arg('--agent-dir=' + agent_dir)
        dumps_dir = self.session.getValue('dst_temporary_directory')
        if dumps_dir is not None:
            cmd.arg('--dumps-dir=' + dumps_dir)
        self.runCmdLine(cmd)

    def getObjectsList(self, agent_name, agent_dir):
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--dump-all')
        cmd.arg('--no-content')
        self.addContext(cmd)
        cmd.arg('--agent-name=' + agent_name)
        cmd.arg('--agent-dir=' + agent_dir)
        self.runCmdLine(cmd)
        
        objects_list = self.makeObjectsListFromDump()
        return objects_list

    def makeObjectsListFromDump(self):
        dump_path = os.path.join(self.getSessionPath(), 'dump.xml')
        
        generator = ObjectsListGenerator()
        
        parser = make_parser()
        parser.setContentHandler(generator)
        parser.parse(dump_path)
        
        return generator.getObjectsList()

    def removeData(self, agent_name, agent_dir):
        if agent_dir == None:
            _logger.info("Agent is not uploaded to source host. Skip removing agent.")
            return
        
        self.session.setValue('step', 'remove-scout')
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--remove-agent')
        self.addContext(cmd)
        cmd.arg('--agent-name=' + agent_name)
        cmd.arg('--agent-dir=' + agent_dir)
        self.runCmdLine(cmd)

    def selectAccounts(self, file_path):
        pass

    def getIpMap(self, file_path):
        dump_path = os.path.join(self.getSessionPath(), 'dump.xml')
        
        generator = IpMappingGenerator(file_path)
        
        parser = make_parser()
        parser.setContentHandler(generator)
        parser.parse(dump_path)
        
        return generator.getIpMapping()

    def clearStatus(self):
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--remove-dumping-status')
        cmd.arg('--session-path=' + self.getSessionPath())
        self.runCmdLine(cmd)

    def getStatusImpl(self, agent_name, agent_dir):
        cmd = self.prepareCmdLine(migration_handler_config.launchpad())
        cmd.arg('--get-dumping-status')
        self.addContext(cmd)
        cmd.arg('--agent-name=' + agent_name)
        cmd.arg('--agent-dir=' + agent_dir)
        
        status = TaskStatus.factory()
        working = Working.factory()
        dumping = WorkingProgress.factory()
        proc = self.runCmdLine(cmd)
        status_output = cmd.get_stdout()
        _logger.debug("Got status: %s" % status_output)
        dumping.build(minidom.parseString(status_output).childNodes[0])
        working.dumping = dumping
        status.working = working

        return status


class ObjectsListGenerator(handler.ContentHandler):
    def __init__(self):
        handler.ContentHandler.__init__(self)
        
        self.objects_list = MigrationObjectList()
        self.current_reseller = None
        self.current_client = None
        self.current_domain = None
        self.current_ip = None
        
        self.is_current_ip_address = False
        self.current_ip_address = ''
        
        self.is_current_ip_type = False
        self.current_ip_type = ''

    def getObjectsList(self):
        return self.objects_list

    # ContentHandler methods

    def startElement(self, name, attrs):
        if name == 'reseller':
            self.current_reseller = ObjTypeReseller(attrs['name'])
            if attrs.has_key('contact'):
                self.current_reseller.set_contact(attrs['contact'])
        elif name == 'client':
            self.current_client = ObjTypeClient(attrs['name'])
            if attrs.has_key('contact'):
                self.current_client.set_contact(attrs['contact'])
        elif name == 'domain':
            self.current_domain = ObjTypeDomain(attrs['name'])
        elif name == 'ip':
            self.current_ip_type = ''
            self.current_ip_address = ''
            self.current_ip = IPMapping()
        elif name == 'ip-type':
            self.current_ip_type = ''
            self.is_current_ip_type = True
        elif name == 'ip-address':
            self.current_ip_address = ''
            self.is_current_ip_address = True

    def endElement(self, name):
        if name == 'reseller':
            self.objects_list.add_reseller(self.current_reseller)
            self.current_reseller = None
        elif name == 'client':
            if self.current_reseller is not None:
                if self.current_reseller.clients is None:
                    self.current_reseller.clients = clients()
                self.current_reseller.clients.add_client(self.current_client)
            else:
                self.objects_list.add_client(self.current_client)
            self.current_client = None
        elif name == 'domain':
            if self.current_client is not None:
                if self.current_client.domains is None:
                    self.current_client.domains = domains()
                self.current_client.domains.add_domain(self.current_domain)
            elif self.current_reseller is not None:
                if self.current_reseller.domains is None:
                    self.current_reseller.domains = domains()
                self.current_reseller.domains.add_domain(self.current_domain)
            else:
                self.objects_list.add_domain(self.current_domain)
            self.current_domain = None
        elif name == 'ip-address':
            self.is_current_ip_address = False
        elif name == 'ip-type':
            self.is_current_ip_type = False
        elif name == 'ip':
            ip_pair = IpPairType( source = self.current_ip_address, destination = None)
            if self.current_ip_type == 'shared':
                self.current_ip.add_shared(ip_pair)
            elif self.current_ip_type == 'exclusive':
                self.current_ip.add_exclusive(ip_pair)
            if self.current_domain is not None:
                self.current_domain.add_ip(self.current_ip)
            elif self.current_client is not None:
                self.current_client.add_ip(self.current_ip)
            elif self.current_reseller is not None:
                self.current_reseller.add_ip(self.current_ip)
            self.current_ip = None
            self.current_ip_type = ''
            self.current_ip_address = ''

    def characters(self, content):
        if self.is_current_ip_type:
            self.current_ip_type += content
        elif self.is_current_ip_address:
            self.current_ip_address += content


class SelectedObjectsGenerator(handler.ContentHandler):
    def __init__(self):
        handler.ContentHandler.__init__(self)
        
        self.current_object_type = False
        self.is_current_name = False
        self.current_name = ''
        self.filter = None
        self.compression_level = None
        
        self.resellers_list = []
        self.clients_list = []
        self.domains_list = []

    def getResellersList(self):
        return self.resellers_list

    def getClientsList(self):
        return self.clients_list

    def getDomainsList(self):
        return self.domains_list

    # ContentHandler methods

    def startElement(self, name, attrs):
        if name == 'reseller':
            self.current_object_type = 'reseller'
        
        elif name == 'client':
            self.current_object_type = 'client'
        
        elif name == 'domain':
            self.current_object_type = 'domain'
        
        elif name == 'name':
            self.is_current_name = True
        
        elif name == 'parameters':
            if "filter" in attrs.keys():
                self.filter = attrs["filter"]
            if "compression-level" in attrs.keys():
                self.compression_level = attrs["compression-level"]

    def endElement(self, name):
        if name == 'reseller':
            self.current_object_type = None
        
        elif name == 'client':
            self.current_object_type = None
        
        elif name == 'domain':
            self.current_object_type = None
        
        elif name == 'name':
            if self.current_object_type == 'reseller':
                self.resellers_list.append(self.current_name)
            elif  self.current_object_type == 'client':
                self.clients_list.append(self.current_name)
            elif  self.current_object_type == 'domain':
                self.domains_list.append(self.current_name)
            
            self.is_current_name = False
            self.current_name = ''

    def characters(self, content):
        if self.is_current_name:
            self.current_name += content


class IpMappingGenerator(handler.ContentHandler):
    def __init__(self, objects_path):
        handler.ContentHandler.__init__(self)
        
        generator = SelectedObjectsGenerator()
        parser = make_parser()
        parser.setContentHandler(generator)
        parser.parse(objects_path)
        
        self.selected_resellers = generator.getResellersList()
        self.selected_clients = generator.getClientsList()
        self.selected_domains = generator.getDomainsList()
        
        self.ip_mapping = IPMapping()
        
        self.is_current_ip = False
        self.current_ip = ''
        self.is_current_ip_type = False
        self.current_ip_type = ''
        self.added_ips = []
        
        self.current_object = None
        self.current_object_type = None

    def getIpMapping(self):
        return self.ip_mapping

    # ContentHandler methods

    def startElement(self, name, attrs):
        if name == 'reseller':
            self.current_object = attrs['name']
            self.current_object_type = 'reseller'
        
        elif name == 'client':
            self.current_object = attrs['name']
            self.current_object_type = 'client'
        
        elif name == 'domain':
            self.current_object = attrs['name']
            self.current_object_type = 'domain'
        
        elif name == 'ip-address':
            self.is_current_ip = True
            self.current_ip = ''
        
        elif name == 'ip-type':
            self.is_current_ip_type = True
            self.current_ip_type = ''

    def endElement(self, name):
        if name == 'reseller':
            self.current_object = None
            self.current_object_type = None
        
        elif name == 'client':
            self.current_object = None
            self.current_object_type = None
        
        elif name == 'domain':
            self.current_object = None
            self.current_object_type = None
        
        elif name == 'ip-address':
            self.is_current_ip = False
        
        elif name == 'ip-type':
            self.is_current_ip_type = False
        
        elif name == 'ip':
            if self.added_ips.count(self.current_ip) == 0:
                objects_list = None
                
                if self.current_object_type == 'reseller':
                    objects_list = self.selected_resellers
                elif  self.current_object_type == 'client':
                    objects_list = self.selected_clients
                elif  self.current_object_type == 'domain':
                    objects_list = self.selected_domains
                
                if objects_list != None and objects_list.count(self.current_object) > 0:
                    ip = IpPairType.factory()
                    ip.set_source(self.current_ip)
                
                    if self.current_ip_type == 'shared':
                        self.ip_mapping.add_shared(ip)
                    elif self.current_ip_type == 'exclusive':
                        self.ip_mapping.add_exclusive(ip)
                    else:
                        raise Exception("Unexpected ip type: " + self.current_ip_type + " (ip " + self.current_ip + ")")
                    
                    self.added_ips.append(self.current_ip)
            
            self.current_ip = ''
            self.is_current_ip_type = ''
            self.is_current_ip = False
            self.is_current_ip_type = False

    def characters(self, content):
        if self.is_current_ip:
            self.current_ip += content
        if self.is_current_ip_type:
            self.current_ip_type += content

